/***************************************************************************
 *            zipitgfx_sample.c
 *
 *  Wed Mar 29 00:20:33 2006
 *  Copyright  2006  Albert Santoni
 *  gamegod [at] sympatico |dot| ca
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include <signal.h>	// for signal()
#include <sys/time.h>	// for setitimer()	
#include <sys/fcntl.h>	// for fcntl()
#include <sys/ioctl.h>  // for ioctl()
#include <termios.h>
#include <unistd.h>

#include "zipitgfx.h"

#define ESCAPE_KEY 0x1B

int event_onInput = 0;
int event_onTimer = 0;
int event_onExit = 0;

//This "trigger" gets called when the program receives the SIGIO signal. (ie. keyboard input callback)
void trigger_Input(int signum)
{
	event_onInput = 1;
	return;
}

//This "trigger" gets called when the program receives the SIGALRM signal. (ie. this is a timer callback.)
void trigger_Timer(int signum)
{
	event_onTimer = 1;
	return;
}

//This "trigger" gets called when the program receives the exit(?) signal.
// [but no args are passed ot it]
// void trigger_Exit(int signum)
void trigger_Exit(void)
{
	event_onExit = 1;
	return;
}

struct itimerval* initTimer()
{
	// install our signal-handler 
	signal(SIGALRM, trigger_Timer);

	// initialize and install our interval-timer
	struct itimerval ourit;
	struct itimerval* oldit = malloc(sizeof(struct itimerval));
		
	// specify the initial delay (i.e., one second)
	ourit.it_value.tv_sec = 1;
	ourit.it_value.tv_usec = 0;
	// specify the periodic delay (i.e., one-half second)
	ourit.it_interval.tv_sec = 0;
	ourit.it_interval.tv_usec = 50000;
	// replace the default itimer with our itimer
	setitimer(ITIMER_REAL, &ourit, oldit); 	
	
	return oldit;
}

//Unregisters the timer signal.
void shutdownTimer(struct itimerval* oldTimer)
{
	// restore original interval-timer and signal-handler
	setitimer(ITIMER_REAL, oldTimer, NULL);
	signal(SIGALRM, SIG_DFL);
	
	return;
}

//This function initializes keyboard input using "termios".
//(It seems to be a little nicer than curses for now.)
struct termios* initKeyboardTermios()
{
	struct termios*	oldTermios = malloc(sizeof(struct termios));
		
	//Enable noncanonical keyboard input
	tcgetattr(STDIN_FILENO, oldTermios);
	struct termios work_termios = *oldTermios;
	work_termios.c_lflag &= ~(ICANON|ECHO|ISIG);
	work_termios.c_cc[ VMIN ] = 1;
	work_termios.c_cc[ VTIME ] = 0;
	tcsetattr( STDIN_FILENO, TCSAFLUSH, &work_termios );
	
	//Activate asynchronous input notification
	signal( SIGIO, trigger_Input);
	fcntl( STDIN_FILENO, F_SETOWN, getpid() );
	int	val = fcntl( STDIN_FILENO, F_GETFL, 0 );
	fcntl( STDIN_FILENO, F_SETFL, val | O_ASYNC );

	return oldTermios;
}

//Shuts down the termios keyboard input.
void shutdownKeyboard(struct termios* oldTermios)
{	
	tcsetattr(STDIN_FILENO, TCSAFLUSH, oldTermios);
	return;
}

// return time between two time samples
int elapsedTime(struct timeval start, struct timeval end)
{
	return (int)((end.tv_sec * 1000 + end.tv_usec / 1000) - (start.tv_sec * 1000 + start.tv_usec / 1000));
}

int main()
{
	int i = 0, x = 0, y = 0;
	float t = 0;
	int	inputChar = 0;
	struct itimerval* oldTimerVal = NULL;
	struct termios* oldTermios = NULL;
	
	struct ScreenStuff* screen = malloc(sizeof(struct ScreenData)); //Allocate space for our framebuffer/framebuffer related stuff.
   	struct ImageData* bmpTest = loadBitmap("logo.bmp"); //Load a test bitmap.
   		
   	//Initialize the framebuffer and clear the screen.
	initScreen(screen); 
	oldTimerVal = initTimer(); //Initialize the rendering timer.
	oldTermios = initKeyboardTermios();  //Initialize keyboard input.
	
	//Register a function that gets called on exit...
	//(Ensures everything gets uninitialized back properly.)
	atexit(trigger_Exit); 	
	
	
	//This is the render loop of the program.
	while (1 == 1)
	{	
		if (event_onTimer)
		{
			i++;
			
			//Draw a funky dancing line on the screen...
			drawLine(screen, 160 + floor(40 * sin(i * 3.14159/16)), 
					 		 160 + floor(40 * cos(i * 3.14159/20)),
					   		 160 + floor(100 * sin(i * 3.14159/20)), 
							 i, 0xF);	
			
			//Draw some text! :)
			drawChar(screen, 102, 140, 30, 0xF); 
			drawText(screen, "Zipit Graphics!!!!!! :)", 40, 100, 0xF);	
			
			//Draw a simple static line. (Use "w/a/s/d" to move the line around)
			drawLine(screen, x, y, 100, 100, 0xF);

			//Create a filled rectange.
			fillRect(screen, 50, 180, 300, 220, 0xF);	
 			
			//Draw a bitmap on the screen.
			if (bmpTest != NULL)
				drawBitmap(screen, bmpTest, 240, 0);
						
			render(screen); //Render the backbuffer (copy it to the screen)

			clearBackBuffer(screen); //Clear our backbuffer.
			
			if (i > 240)
				i = 0;
			
			event_onTimer = 0;
			
		}
		
		//Termios-based keyboard input (non-curses)...	
		if (event_onInput)
		{
			inputChar = 0;
			read(STDIN_FILENO, &inputChar, 4);
			
			if (inputChar == ESCAPE_KEY) 
			{
				read(STDIN_FILENO, &inputChar, 4);
				//printf( "\'%c\'\n", inputChar);
				read(STDIN_FILENO, &inputChar, 4);
				//printf( "\'%c\'\n", inputChar);			
				event_onExit = 1;
			}
			//if (inputChar == 0x61)
			//	i = 0;			
			if (inputChar == 'w')
				y -= 5;
			else if (inputChar == 'a')
				x -= 5;
			else if (inputChar == 's')
				y += 5;
			else if (inputChar == 'd')
				x += 5;
			//else
			//	printf("Unbound key: \'%c\'\n", inputChar);
			
			event_onInput = 0; //Reset the "input" trigger.
			usleep(50000); //Sleep a little after input (input goes crazy if we don't?)
		}
		
		if (event_onExit)
		{
			shutdownScreen(screen);
			shutdownTimer(oldTimerVal);
			shutdownKeyboard(oldTermios);
			event_onExit = 0; //Reset the "exit" trigger. (Kinda pointless...)
			printf("Successfully exited.\n");
			exit(0);
		}
		
		//Curses based key input...
	/*	switch (getch()) {
			case KEY_UP:
				//printw("UP\n");
			break;
			case KEY_DOWN:
				//printw("DOWN\n");
				break;
			case KEY_LEFT:
				//printw("LEFT\n");
				break;
			case KEY_RIGHT:
				//printw("RIGHT\n");
				break;
			case 'q':
				event_onExit = 1;
				break;
		}

		endwin();		
		*/
	}
}
